﻿# coding:utf-8
# --------------------------------------------------------------------
# 名称:       流線図の作成 ツール(CreateFlowline tool)
# 作成日:     2020年2月4日
# Copyright:  WINGFIELD since1981
# 説明:       ライン フィーチャを指定したパラメーターで曲線に変換する
#             投影座標系のみサポート
# 操作方法:   下記URLを参照
#             https://www.wingfield.gr.jp/archives/11363
#
# 履歴:
#             2020-02-04  新規作成
#             2020-02-11  パラメーター ELLIPTIC, BEZIER6 を追加
#                         直線以外のセグメント使用時のエラー制御を追加
#             2020-02-12  係数フィールドの未指定に対応
#                         use_quadran のデータ タイプを変更
#                         BEZIER5,6 でコントロール ポイントが正しく
#                         設定されない問題への対応
#             2020-02-13  ジオメトリ フィールドが2列目以外に存在する場合を考慮
#                         arc_type に GEODESIC（測地線）を追加
# --------------------------------------------------------------------

import arcpy
import math
import json
import numpy
import sys
import os
import distutils.util
import ast

try:
  ## 関数
  # フィールドのインデックス番号取得（未取得時は例外）
  def findindex(table, fieldname):
    return [i.name for i in arcpy.ListFields(table)].index(fieldname)

  # --------------------------------------------------------------------
  # 処理
  # --------------------------------------------------------------------
  # 入力パラメーター
  input_features      = arcpy.GetParameterAsText(0)   # 入力フィーチャ
  output_featureclass = arcpy.GetParameterAsText(1)   # 出力フィーチャクラス
  arc_type            = arcpy.GetParameterAsText(2)   # 曲線タイプ（LINE, GEODESIC, CIRCULAR, ELLIPTIC, BEZIER1, BEZIER2, BEZIER3, BEZIER4, BEZIER5, BEZIER6）
  factor_field        = arcpy.GetParameterAsText(3)   # 係数フィールド
  factor_value        = arcpy.GetParameterAsText(4)   # 係数 (1/f)
  use_quadrant        = arcpy.GetParameterAsText(5)   # 4象限に応じた曲線方向の設定
  offset_x            = arcpy.GetParameterAsText(6)   # オフセットX距離
  offset_y            = arcpy.GetParameterAsText(7)   # オフセットY距離
  use_densify         = arcpy.GetParameterAsText(8)   # 頂点を挿入

  # --------------------------------------
  # 実行
  # --------------------------------------
  arcpy.AddMessage('処理開始...')
  # パラメーター確認
  # 入力フィーチャ
  if input_features == None or input_features == '':
    input_features = r""

  # 出力フィーチャクラス
  if output_featureclass == None or output_featureclass == '':
    output_featureclass = r""

  # 曲線タイプ  LINE, GEODESIC, CIRCULAR, ELLIPTIC, BEZIER1, BEZIER2, BEZIER3, BEZIER4, BEZIER5, BEZIER6
  if arc_type == None or arc_type == '':
    arc_type = 'CIRCULAR'

  # 係数フィールド
  if factor_field == None or factor_field == '':
    factor_field = r''

  # 係数 (1/f)
  if factor_value == None or factor_value == '':
    factor_value = '5.0'
  factor_value = float(factor_value)

  # 4象限に応じた曲線方向の設定 'NONE':未使用, 'X':X軸に適用, 'Y':Y軸に適用
  if use_quadrant  == None or use_quadrant == '':
    use_quadrant = 'NONE'

  # 原点オフセットX座標
  if offset_x  == None or offset_x == '':
    offset_x = '0.0'
  offset_x = float(offset_x)

  # 原点負オフセットY座標
  if offset_y  == None or offset_y == '':
    offset_y = '0.0'
  offset_y = float(offset_y)

  # 頂点を挿入
  if use_densify  == None or use_densify == '':
    use_densify = 'true'
  use_densify = distutils.util.strtobool(use_densify)

  # 固定値
  alias_name = '' #出力フィーチャクラスのエイリアス名

  # Describe オブジェクトの取得
  input_describe = arcpy.Describe(input_features)

  # フィールド名の制御
  fields = arcpy.ListFields(input_features)
  
  #入力フィールドの有無を確認
  shape_field_name = input_describe.shapeFieldName
  shape_field_index = findindex(input_features,shape_field_name)

  if factor_field == '':
    factor_field_index = -1
  else:
    if arcpy.ListFields(input_features,factor_field):
      factor_field_index = findindex(input_features, factor_field)
    else:
      arcpy.AddError('指定したフィールドは存在しません。')
      sys.exit()
  
  new_fields = []
  for field in fields:
    if field.type == 'Geometry':
      new_fields.append('SHAPE@JSON')
    else:
      new_fields.append(field.name)

  # 前提条件の確認
  if input_describe.shapeType != 'Polyline':
    arcpy.AddError('ポリライン以外は処理できません。')
    sys.exit()

  if input_describe.spatialReference.PCSName == '':
    arcpy.AddError('地理座標系は処理できません。')
    sys.exit()

  # 出力フィーチャクラス
  workspace_name = os.path.dirname(output_featureclass)
  featureclass_name = os.path.basename(output_featureclass)
  output_featureclass = arcpy.management.CreateFeatureclass(workspace_name,featureclass_name, 'POLYLINE', input_features, 'DISABLED', 'DISABLED', input_describe.spatialReference, None, 0, 0, 0, alias_name)

  # フィーチャの作成
  create_values = []
  search_cursor = arcpy.da.SearchCursor(input_features, new_fields)

  for row in search_cursor:
    polyline = row[shape_field_index]     # Shape フィールド
    
    if polyline is None:  # 距離が0の場合
      continue
    
    # ジオメトリの取得
    # 計算
    polyline_json = json.loads(polyline)
    if polyline_json.get('paths') is None:
      arcpy.AddError('入力フィーチャのジオメトリは直線以外のセグメントをサポートしていません。')
      sys.exit()
    point_from = polyline_json['paths'][0][0]
    point_to = polyline_json['paths'][0][-1]
    sr = polyline_json['spatialReference']
    x1 = point_from[0]
    y1 = point_from[1]
    x2 = point_to[0]
    y2 = point_to[1]

    # 4象限による曲線方向の決定
    if use_quadrant == 'NONE':
      plusminus = 1 
    elif use_quadrant == 'X':
      # 4象限による曲線方向の決定（X軸を考慮）
      if x1 <= x2 - offset_x and y1 <= y2 - offset_y: # 第1象限
        plusminus = -1
      if x1 > x2 - offset_x and y1 <= y2 - offset_y:  # 第2象限
        plusminus = 1
      if x1 <= x2 - offset_x and y1 > y2 - offset_y:  # 第3象限
        plusminus = 1
      if x1 > x2 - offset_x and y1 > y2 - offset_y:  # 第4象限
        plusminus = -1
    elif use_quadrant == 'Y':
      if x1 <= x2 - offset_x and y1 <= y2 - offset_y: # 第1象限
        plusminus = 1
      if x1 > x2 - offset_x and y1 <= y2 - offset_y:  # 第2象限
        plusminus = -1
      if x1 <= x2 - offset_x and y1 > y2 - offset_y:  # 第3象限
        plusminus = -1
      if x1 > x2 - offset_x and y1 > y2 - offset_y:  # 第4象限
        plusminus = 1
    else:
      arcpy.AddError('パラメーターが不正です（4象限に応じた曲線方向の設定）')
      sys.exit()

    # 係数フィールドが未指定の場合
    if factor_field_index == -1:
      len_s = 1 / factor_value * plusminus             # 線分からの垂直距離
    else:
      len_s = row[factor_field_index] / factor_value * plusminus             # 線分からの垂直距離

    len_l = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)     # 線分長
    x3 = (x1 + x2) / 2.0                                  # 線分中心X座標
    y3 = (y1 + y2) / 2.0                                  # 線分中心Y座標

    # 線分角度
    rad = math.atan2(y2 - y1, x2 - x1) - math.radians(90) # 線分の垂直線の角度

    x4 = x3 + len_s * math.cos(rad)                       # 離心X座標
    y4 = y3 + len_s * math.sin(rad)                       # 離心Y座標

    # ベジェ曲線用コントロール ポイント
    x1b = x1 + len_s * math.cos(rad)                      # 始点X座標用コントロール ポイント
    y1b = y1 + len_s * math.sin(rad)                      # 始点Y座標用コントロール ポイント
    x2b = x2 + len_s * math.cos(rad)                      # 終点X座標用コントロール ポイント
    y2b = y2 + len_s * math.sin(rad)                      # 終点Y座標用コントロール ポイント
    
    # ベジェ曲線による楕円弧用コントール ポイント
    elliptic_factor = math.radians(-45 * plusminus * numpy.sign(factor_value))
    x1e = x1 + len_s * math.cos(rad - elliptic_factor)  # 始点X座標用コントロール ポイント
    y1e = y1 + len_s * math.sin(rad - elliptic_factor)  # 始点Y座標用コントロール ポイント
    x2e = x2 + len_s * math.cos(rad + elliptic_factor)  # 終点X座標用コントロール ポイント
    y2e = y2 + len_s * math.sin(rad + elliptic_factor)  # 終点Y座標用コントロール ポイント

    # JSON形式サンプル
    # https://developers.arcgis.com/documentation/common-data-types/geometry-objects.htm
    # Line          {'paths':[[[0,0],[1,1]]]}
    # CircularArc   {'curvePaths':[[[0,0],{'c':[[1,1],[5:5]]}]]}
    # EllipticArc   {'curvePaths':[[[0,0],{'a':[[1,1],[5,5],0,0,0,0,0]}]]} #{'curvePaths':[[[0,0],{'a':[[toX,toY],[centerX,centerY],minor(0or1),clockwise(0or1),rotation(radians),semi-major axis,ratio]}]]}
    # BezierArc     {'curvePaths':[[[0,0],{'b':[[1,1],[1,0],[0,1]]}]]}

    # JSON 構築
    if arc_type == 'LINE':
      # 直線
      json_str = "{{'paths':[[[{0},{1}],[{2},{3}]]]}}".format(x1,y1,x2,y2)
    elif arc_type == 'GEODESIC':
      # 測地線
      json_str = "{{'paths':[[[{0},{1}],[{2},{3}]]],'spatialReference':{4}}}".format(x1,y1,x2,y2,json.dumps(sr))
      poyline_json = ast.literal_eval(json_str)
      polyline = arcpy.AsShape(polyline_json, True)
      polyline_geodesic = polyline.densify('GEODESIC', 10000.0, 1.0)  # 10km で Densify
      json_str = polyline_geodesic.JSON

    elif arc_type == 'CIRCULAR':
      # 円弧
      json_str = "{{'curvePaths':[[[{0},{1}],{{'c':[[{2},{3}],[{4},{5}]]}}]]}}".format(x1,y1,x2,y2,x4,y4)
    
    elif arc_type == 'ELLIPTIC':
      # 楕円弧（ベジェ曲線から作成）
      x5 = x3 + plusminus * (y1 - y2) / factor_value   # 楕円の中心X座標
      y5 = y3 - plusminus * (x1 - x2) / factor_value   # 楕円の中心Y座標
      e_minor = 1 #短長
      if plusminus == 1:
        e_clockwise = 0                         # 楕円の向き
      else:
        e_clockwise = 1
      e_rotation = math.atan2(y2 - y1, x2 - x1) # 楕円の角度
      e_axis = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2) / math.sqrt(3)  # 楕円の長半径
      e_factor_value = 2 * math.sqrt(3) / factor_value                    # 偏平率
      json_str = "{{'curvePaths':[[[{0},{1}],{{'a':[[{2},{3}],[{4},{5}],{6},{7},{8},{9},{10}]}}]]}}".format(x1,y1,x2,y2,x5,y5,e_minor,e_clockwise,e_rotation,e_axis,e_factor_value)
    
    elif arc_type == 'BEZIER1':
      # ベジェ曲線1（始点終点からの垂線上）
      json_str = "{{'curvepaths':[[[{0},{1}],{{'b':[[{2},{3}],[{4},{5}],[{6},{7}]]}}]]}}".format(x1,y1,x2,y2,x1b,y1b,x2b,y2b)
    
    elif arc_type == 'BEZIER2':
      # ベジェ曲線2（中間点からの垂線上）
      json_str = "{{'curvepaths':[[[{0},{1}],{{'b':[[{2},{3}],[{4},{5}],[{6},{7}]]}}]]}}".format(x1,y1,x2,y2,x4,y4,x4,y4)
    
    elif arc_type == 'BEZIER3':
      # ベジェ曲線3（始点のみ垂線上）
      json_str = "{{'curvepaths':[[[{0},{1}],{{'b':[[{2},{3}],[{4},{5}],[{6},{7}]]}}]]}}".format(x1,y1,x2,y2,x1b,y1b,x4,y4)
    
    elif arc_type == 'BEZIER4':
      # ベジェ曲線4（終点のみ垂線上）
      json_str = "{{'curvepaths':[[[{0},{1}],{{'b':[[{2},{3}],[{4},{5}],[{6},{7}]]}}]]}}".format(x1,y1,x2,y2,x4,y4,x2b,y2b)
    
    elif arc_type == 'BEZIER5':
      # ベジェ曲線5（楕円弧ライク）
      json_str = "{{'curvepaths':[[[{0},{1}],{{'b':[[{2},{3}],[{4},{5}],[{6},{7}]]}}]]}}".format(x1,y1,x2,y2,x1e,y1e,x2e,y2e)
    
    elif arc_type == 'BEZIER6':
      # ベジェ曲線6（楕円弧ライク）
      # コントロール ポイントのX座標が象限を超えないように対処
      if use_quadrant == 'X':
        if x1 < x2:
          if x1e < x1:
            x1e = x1
          if x2e < x1:
            x2e = x1
        if x2 < x1:
          if x1e > x1:
            x1e = x1
          if x2e > x1:
            x2e = x1
        else:         # x2 == x1
            x1e = x1
            x2e = x1
      elif use_quadrant == 'Y':
        if y1 < y2:
          if y1e < y1:
            y1e = y1
          if y2e < y1:
            y2e = y1
        if y2 < y1:
          if y1e > y1:
            y1e = y1
          if y2e > y1:
            y2e = y1
        else:         # y2 == y1
            y1e = y1
            y2e = y1
      else:
        arcpy.AddError('このパラメーターは使用できません。arc_type が BEZIER6 の場合は use_quadrant は NONE 以外を指定してください。')
        sys.exit()
      json_str = "{{'curvepaths':[[[{0},{1}],{{'b':[[{2},{3}],[{4},{5}],[{6},{7}]]}}]]}}".format(x1,y1,x2,y2,x1e,y1e,x2e,y2e)
    new_row = []
    for i in range(len(row)):
      if i == shape_field_index:
        new_row.append(json_str)
      else:
        new_row.append(row[i])


    # カーブの作成
    create_value = tuple(new_row)

    # リストへの統合
    create_values.append(create_value)

  # タプルへの変換
  create_values = tuple(create_values)

  # インサート カーソル
  insert_cursor = arcpy.da.InsertCursor(output_featureclass, new_fields)
  for row in create_values:
    insert_cursor.insertRow(row)

  # オブジェクト参照を解放
  del row, search_cursor, insert_cursor

  # 頂点を挿入
  if use_densify == True:
    arcpy.edit.Densify(output_featureclass, "OFFSET", "10 Unknown", "1 Unknown", 1)

  arcpy.AddMessage('処理が完了しました。')

except arcpy.ExecuteError as e:
  arcpy.AddError('エラーが発生しました。')
  arcpy.AddError(str(e))

  # オブジェクト参照を解放
  # del search_cursor, insert_cursor
